@@ -4,55 +4,73 @@
 This set of utlities can be extended but must be backward compatible for at
 least two versions
 """
+import copy
 import datetime
 import importlib
 import json
 import logging
 import math
+import os
 import pathlib
+import pickle
+import random
+import re
 import string
 import sys
-from typing import List, Optional, Iterable, Union, Callable, Mapping, Any, Sequence, Tuple, Dict, Type
-from typing import TYPE_CHECKING
+from enum import Enum
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    List,
+    Mapping,
+    Optional,
+    Sequence,
+    Tuple,
+    Type,
+    Union,
+)
+
+import colorlog
+import inflect
+import numpy as np
+import pandas as pd
+import scipy.stats as stats
+import stringcase
+import yaml
+
+from negmas.config import NEGMAS_CONFIG
+from negmas.generics import *
 
 if TYPE_CHECKING:
     pass
 
-import colorlog
-import numpy as np
-import os
-import random
-import re
-import scipy.stats as stats
-import inflect
-import stringcase
-from enum import Enum
-import yaml
-from negmas.generics import *
-import copy
-
 __all__ = [
-    'create_loggers',
+    "shortest_unique_names",
+    "create_loggers",
     # 'MultiIssueUtilityFunctionMapping',
-    'ReturnCause',
-    # The cause for returning from a sync call
-    # 'LazyInitializable',    # A class that allows setting initialization parameters later
-    # via a set_params() call
-    'LoggerMixin',
-    # Adds the ability to log to screen and file
-    # The base for all named _entities in the system
-    'Distribution',  # A probability distribution
-    'snake_case',
-    'camel_case',
-    'unique_name',
-    'is_nonzero_file',
-    'pretty_string',
-    'ConfigReader',
-    'get_class',
-    'import_by_name',
-    'get_full_type_name',
-    'instantiate',
-    'humanize_time'
+    "ReturnCause",
+    "Distribution",  # A probability distribution
+    "snake_case",
+    "camel_case",
+    "unique_name",
+    "is_nonzero_file",
+    "pretty_string",
+    "ConfigReader",
+    "get_class",
+    "import_by_name",
+    "get_full_type_name",
+    "instantiate",
+    "humanize_time",
+    "gmap",
+    "ikeys",
+    "Floats",
+    "DEFAULT_DUMP_EXTENSION",
+    "dump",
+    "add_records",
+    "load",
 ]
 # conveniently named classes
 
@@ -67,11 +85,15 @@ IterableMappings = List[IterableMapping]
 ParamLists = Iterable[ParamList]
 Floats = List[float]
 
-COMMON_LOG_FILE_NAME = './logs/{}_{}.txt'.format('log', datetime.datetime.now().strftime('%Y%m%d-%H%M%S'))
+COMMON_LOG_FILE_NAME = "./logs/{}_{}.txt".format(
+    "log", datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
+)
 
 MODULE_LOG_FILE_NAME: Dict[str, str] = dict()
 
-LOGS_BASE_DIR = './logs'
+LOGS_BASE_DIR = "./logs"
+
+DEFAULT_DUMP_EXTENSION = NEGMAS_CONFIG.get("default_dump_extension", "json")
 
 
 class ReturnCause(Enum):
@@ -80,12 +102,46 @@ class ReturnCause(Enum):
     FAILURE = 2
 
 
+def shortest_unique_names(strs: List[str], sep="."):
+    """
+    Finds the shortest unique strings starting from the end of each input
+    string based on the separator.
+
+    The final strings will only be unique if the inputs are unique.
+
+    Example:
+        given ["a.b.c", "d.e.f", "a.d.c"] it will generate ["b.c", "f", "d.c"]
+    """
+    lsts = [_.split(sep) for _ in strs]
+    names = [_[-1] for _ in lsts]
+    if len(names) == len(set(names)):
+        return names
+    locs = defaultdict(list)
+    for i, s in enumerate(names):
+        locs[s].append(i)
+    mapping = {"": ""}
+    for s, l in locs.items():
+        if len(s) < 1:
+            continue
+        if len(l) == 1:
+            mapping[strs[l[0]]] = s
+            continue
+        strs_new = [sep.join(lsts[_][:-1]) for _ in l]
+        prefixes = shortest_unique_names(strs_new, sep)
+        for loc, prefix in zip(l, prefixes):
+            x = sep.join([prefix, s])
+            if x.startswith(sep):
+                x = x[len(sep) :]
+            mapping[strs[loc]] = x
+    return [mapping[_] for _ in strs]
+
+
 def create_loggers(
     file_name: Optional[str] = None,
     module_name: Optional[str] = None,
     screen_level: Optional[int] = logging.WARNING,
     file_level: Optional[int] = logging.DEBUG,
-    format_str: str = '%(asctime)s - %(levelname)s - %(message)s',
+    format_str: str = "%(asctime)s - %(levelname)s - %(message)s",
     colored: bool = True,
     app_wide_log_file: bool = True,
     module_wide_log_file: bool = False,
@@ -114,30 +170,32 @@ def create_loggers(
 
     """
     if module_name is None:
-        module_name = __file__.split('/')[-1][:-3]
+        module_name = __file__.split("/")[-1][:-3]
     # create logger if it does not already exist
-    logger = logging.getLogger(module_name)
-    if len(logger.handlers) > 0:
-        return logger
-    logger.setLevel(logging.DEBUG)
+    logger = None
+    if module_wide_log_file or app_wide_log_file:
+        logger = logging.getLogger(module_name)
+        if len(logger.handlers) > 0:
+            return logger
+        logger.setLevel(logging.DEBUG)
     # create formatter
-    if colored and 'colorlog' in sys.modules and os.isatty(2):
-        date_format = '%Y-%m-%d %H:%M:%S'
-        cformat = '%(log_color)s' + format_str
+    if colored and "colorlog" in sys.modules and os.isatty(2):
+        date_format = "%Y-%m-%d %H:%M:%S"
+        cformat = "%(log_color)s" + format_str
         formatter = colorlog.ColoredFormatter(
             cformat,
             date_format,
             log_colors={
-                'DEBUG': 'reset',
-                'INFO': 'green',
-                'WARNING': 'yellow',
-                'ERROR': 'red',
-                'CRITICAL': 'bold_red',
+                "DEBUG": "reset",
+                "INFO": "green",
+                "WARNING": "yellow",
+                "ERROR": "red",
+                "CRITICAL": "bold_red",
             },
         )
     else:
         formatter = logging.Formatter(format_str)
-    if screen_level is not None:
+    if screen_level is not None and (module_wide_log_file or app_wide_log_file):
         # create console handler and set level to logdebug
         screen_logger = logging.StreamHandler()
         screen_logger.setLevel(screen_level)
@@ -147,19 +205,24 @@ def create_loggers(
         logger.addHandler(screen_logger)
     if file_name is not None and file_level is not None:
         file_name = str(file_name)
+        if logger is None:
+            logger = logging.getLogger(file_name)
+            logger.setLevel(file_level)
         if len(file_name) == 0:
             if app_wide_log_file:
                 file_name = COMMON_LOG_FILE_NAME
             elif module_wide_log_file and module_name in MODULE_LOG_FILE_NAME.keys():
                 file_name = MODULE_LOG_FILE_NAME[module_name]
             else:
-                file_name = '{}/{}_{}.txt'.format(LOGS_BASE_DIR,
-                    module_name, datetime.datetime.now().strftime('%Y%m%d-%H%M%S')
+                file_name = "{}/{}_{}.txt".format(
+                    LOGS_BASE_DIR,
+                    module_name,
+                    datetime.datetime.now().strftime("%Y%m%d-%H%M%S"),
                 )
                 MODULE_LOG_FILE_NAME[module_name] = file_name
 
-            os.makedirs(f'{LOGS_BASE_DIR}', exist_ok=True)
-        os.makedirs(os.path.dirname(file_name), exist_ok=True) # type: ignore
+            os.makedirs(f"{LOGS_BASE_DIR}", exist_ok=True)
+        os.makedirs(os.path.dirname(file_name), exist_ok=True)  # type: ignore
         file_logger = logging.FileHandler(file_name)
         file_logger.setLevel(file_level)
         file_logger.setFormatter(formatter)
@@ -183,14 +246,14 @@ def snake_case(s: str) -> str:
     Returns:
         str: converted string
     """
-    return re.sub(
-        '(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))', '_\\1', s
-    ).lower().strip(
-        '_'
+    return (
+        re.sub("(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))", "_\\1", s).lower().strip("_")
     )
 
 
-def camel_case(s: str, capitalize_first: bool = False, lower_first: bool = False) -> str:
+def camel_case(
+    s: str, capitalize_first: bool = False, lower_first: bool = False
+) -> str:
     """ Converts a string from snake_case to CamelCase
 
     Example:
@@ -214,18 +277,20 @@ def camel_case(s: str, capitalize_first: bool = False, lower_first: bool = False
     """
     if len(s) < 1:
         return s
-    parts = s.split('_')
+    parts = s.split("_")
     if capitalize_first:
         parts = [_.capitalize() for _ in parts]
     elif lower_first:
-        parts = [parts[0].lower()] +  [_.capitalize() for _ in parts[1:]]
+        parts = [parts[0].lower()] + [_.capitalize() for _ in parts[1:]]
     else:
         parts = [parts[0]] + [_.capitalize() for _ in parts[1:]]
 
-    return ''.join(parts)
+    return "".join(parts)
 
 
-def unique_name(base: str, add_time=True, rand_digits=8) -> str:
+def unique_name(
+    base: Union[pathlib.Path, str], add_time=True, rand_digits=8, sep="/"
+) -> str:
     """Return a unique name.
 
     Can be used to return a unique directory name on the givn base.
@@ -239,26 +304,26 @@ def unique_name(base: str, add_time=True, rand_digits=8) -> str:
     Examples:
 
         >>> a = unique_name('')
-        >>> len(a) == 8 + 1 + 6 + 8
+        >>> len(a) == 8 + 1 + 6 + 8 + 6
         True
 
     Returns:
         str: The unique name.
 
     """
+    _time, rand_part = "", ""
     if rand_digits > 0:
-        characters = string.ascii_letters + string.digits
-        password = "".join(
-            random.choice(characters) for _ in range(rand_digits)
+        rand_part = "".join(
+            random.choices(string.digits + string.ascii_letters, k=rand_digits)
         )
-    else:
-        password = ''
     if add_time:
-        _time = datetime.datetime.now().strftime('%Y%m%d-%H%M%S')
-    else:
-        _time = ''
-    sub = _time + password
-    return os.path.join(base, sub) if len(sub) > 0 else base
+        _time = datetime.datetime.now().strftime("%Y%m%dH%H%M%S%f")
+    sub = _time + rand_part
+    if len(sub) == 0:
+        return base
+    if len(base) == 0:
+        return sub
+    return f"{str(base)}{sep}{sub}"
 
 
 def is_nonzero_file(fpath: str) -> bool:
@@ -286,15 +351,15 @@ def pretty_string(src: Any, tab_size=2, compact=False) -> str:
             - This function assumes that the patterns `` "`` and ``":`` do not appear anywhere in the input.
               If they appear, the space, : will be removed.
         """
-    s = _pretty_string(src, dpth=0, current_key='', tab_size=tab_size)
+    s = _pretty_string(src, dpth=0, current_key="", tab_size=tab_size)
     if compact:
-        return s.replace('\n', '')
+        return s.replace("\n", "")
 
     else:
-        return s.replace(' "', ' ').replace('":', ':')
+        return s.replace(' "', " ").replace('":', ":")
 
 
-def _pretty_string(src, dpth=0, current_key='', tab_size=2) -> str:
+def _pretty_string(src, dpth=0, current_key="", tab_size=2) -> str:
     """Recursively print nested elements.
 
     Args:
@@ -307,24 +372,24 @@ def _pretty_string(src, dpth=0, current_key='', tab_size=2) -> str:
     """
 
     def tabs(n):
-        return ' ' * n * tab_size  # or 2 or 8 or...
+        return " " * n * tab_size  # or 2 or 8 or...
 
-    output = ''
+    output = ""
     if isinstance(src, dict):
-        output += tabs(dpth) + '{\n'
+        output += tabs(dpth) + "{\n"
         for key, value in src.items():
-            output += _pretty_string(value, dpth + 1, key) + '\n'
-        output += tabs(dpth) + '}'
+            output += _pretty_string(value, dpth + 1, key) + "\n"
+        output += tabs(dpth) + "}"
     elif isinstance(src, list) or isinstance(src, tuple):
-        output += tabs(dpth) + '[\n'
+        output += tabs(dpth) + "[\n"
         for litem in src:
-            output += _pretty_string(litem, dpth + 1) + '\n'
-        output += tabs(dpth) + ']'
+            output += _pretty_string(litem, dpth + 1) + "\n"
+        output += tabs(dpth) + "]"
     else:
         if len(current_key) > 0:
-            output += (tabs(dpth) + '"%s":%s' % (current_key, src))
+            output += tabs(dpth) + '"%s":%s' % (current_key, src)
         else:
-            output += (tabs(dpth) + '%s' % src)
+            output += tabs(dpth) + "%s" % src
     return output
 
 
@@ -411,87 +476,6 @@ class LazyInitializable(object):
         pass
 
 
-class LoggerMixin(object):
-    """Parent of all agents that can log to the console/file.
-
-    Args:
-        file_name (Optional[str]): File name to use for logging
-
-    Examples:
-        Create a temporary file to test with
-
-        >>> from tempfile import mkstemp
-        >>> _, f_name = mkstemp()
-        >>> l = LoggerMixin(f_name)
-        >>> l.loginfo('test info')
-        >>> l.logerror('test error')
-        >>> l.logwarning('test warning')
-        >>> l.logdebug('test debug')
-        >>> with open(f_name, 'r') as f:
-        ...     lines = f.readlines()
-        >>> ('INFO - test info' in lines[0]) if len(lines) > 0 else True
-        True
-        >>> ('ERROR - test error' in lines[1]) if len(lines) > 1 else True
-        True
-        >>> ('WARNING - test warning' in lines[2])  if len(lines) > 2 else True
-        True
-        >>> ('DEBUG - test debug' in lines[3])  if len(lines) > 3 else True
-        True
-
-    """
-
-    def __init__(self, file_name: Optional[str] = None, screen_log: bool=False) -> None:
-        """Constructor
-
-        Constructs a logger agent
-
-        Args:
-            file_name (str, optional): Defaults to None. File used for
-
-        """
-        self.log_file_name = file_name
-        if screen_log:
-            self.logger = create_loggers(self.log_file_name)
-        else:
-            self.logger = create_loggers(self.log_file_name, screen_level=None)
-
-    def loginfo(self, s: str) -> None:
-        """logs info-level information
-
-        Args:
-            s (str): The string to log
-
-        """
-        self.logger.info(s)
-
-    def logdebug(self, s) -> None:
-        """logs debug-level information
-
-        Args:
-            s (str): The string to log
-
-        """
-        self.logger.debug(s)
-
-    def logwarning(self, s) -> None:
-        """logs warning-level information
-
-        Args:
-            s (str): The string to log
-
-        """
-        self.logger.warning(s)
-
-    def logerror(self, s) -> None:
-        """logs error-level information
-
-        Args:
-            s (str): The string to log
-
-        """
-        self.logger.error(s)
-
-
 class Distribution(object):
     """Any distribution from scipy.stats with overloading of addition and multiplication.
 
@@ -517,27 +501,27 @@ class Distribution(object):
 
     """
 
-    def __init__(
-        self,
-        dtype: str,
-        **kwargs,
-    ) -> None:
+    def __init__(self, dtype: str, **kwargs) -> None:
         super().__init__()
         dist = getattr(stats, dtype.lower(), None)
         if dist is None:
-            raise ValueError(f'Unknown distribution {dtype}')
-        if 'loc' not in kwargs.keys():
-            kwargs['loc'] = 0.0
-        if 'scale' not in kwargs.keys():
-            kwargs['scale'] = 1.0
+            raise ValueError(f"Unknown distribution {dtype}")
+        if "loc" not in kwargs.keys():
+            kwargs["loc"] = 0.0
+        if "scale" not in kwargs.keys():
+            kwargs["scale"] = 1.0
 
         self.dist = dist(**kwargs)
         self.dtype = dtype
         self.__cached = None
 
     @classmethod
-    def around(cls, value: float = 0.5, range: Tuple[float, float] = (0.0, 1.0),
-               uncertainty: float = 0.5) -> 'Distribution':
+    def around(
+        cls,
+        value: float = 0.5,
+        range: Tuple[float, float] = (0.0, 1.0),
+        uncertainty: float = 0.5,
+    ) -> "Distribution":
         """
         Generates a uniform distribution around the input value in the given range with given uncertainty
 
@@ -550,18 +534,20 @@ class Distribution(object):
             Distribution A uniform distribution around `value` with uncertainty (scale) `uncertainty`
         """
         if uncertainty >= 1.0:
-            return cls(dtype='uniform', loc=range[0], scale=range[1])
+            return cls(dtype="uniform", loc=range[0], scale=range[1])
         if uncertainty <= 0.0:
-            return cls(dtype='uniform', loc=value, scale=0.0)
+            return cls(dtype="uniform", loc=value, scale=0.0)
         scale = uncertainty * (range[1] - range[0])
         loc = max(range[0], (random.random() - 1.0) * scale + value)
         if loc + scale > range[1]:
-            loc -= (loc + scale - range[1])
-        return cls(dtype='uniform', loc=loc, scale=scale)
+            loc -= loc + scale - range[1]
+        return cls(dtype="uniform", loc=loc, scale=scale)
 
     def mean(self) -> float:
-        if self.dtype != 'uniform':
-            raise NotImplementedError('Only uniform distributions are supported for now')
+        if self.dtype != "uniform":
+            raise NotImplementedError(
+                "Only uniform distributions are supported for now"
+            )
         if self.scale < 1e-6:
             return self.loc
         mymean = self.dist.mean()
@@ -573,26 +559,22 @@ class Distribution(object):
     def __and__(self, other):
         if isinstance(other, float) or isinstance(other, int):
             return float(other)
-        if self.dtype == 'uniform':
+        if self.dtype == "uniform":
             beg = max(self.loc, other.loc)
             end = min(self.scale + self.loc, other.loc + other.scale)
-            return Distribution(self.dtype
-                                , loc=beg
-                                , scale=end - beg
-                                )
+            return Distribution(self.dtype, loc=beg, scale=end - beg)
         raise NotImplementedError()
 
     def __or__(self, other):
         if isinstance(other, float) or isinstance(other, int):
             return float(other)
-        if self.dtype == 'uniform':
-            raise NotImplementedError('Current implementation assumes an overlap otherwise a mixture must be returned')
+        if self.dtype == "uniform":
+            raise NotImplementedError(
+                "Current implementation assumes an overlap otherwise a mixture must be returned"
+            )
             beg = min(self.loc, other.loc)
             end = max(self.scale + self.loc, other.loc + other.scale)
-            return Distribution(self.dtype
-                                , loc=beg
-                                , scale=end - beg
-                                )
+            return Distribution(self.dtype, loc=beg, scale=end - beg)
         raise NotImplementedError()
 
     def prob(self, val: float) -> float:
@@ -605,16 +587,16 @@ class Distribution(object):
 
     @property
     def loc(self):
-        return self.dist.kwds.get('loc', 0.0)
+        return self.dist.kwds.get("loc", 0.0)
 
     @property
     def scale(self):
-        return self.dist.kwds.get('scale', 0.0)
+        return self.dist.kwds.get("scale", 0.0)
 
     def __str__(self):
-        if self.dtype == 'uniform':
-            return f'U({self.loc}, {self.loc+self.scale})'
-        return f'{self.dtype}(loc:{self.loc}, scale:{self.scale})'
+        if self.dtype == "uniform":
+            return f"U({self.loc}, {self.loc+self.scale})"
+        return f"{self.dtype}(loc:{self.loc}, scale:{self.scale})"
 
     def __copy__(self):
         cls = self.__class__
@@ -632,6 +614,24 @@ class Distribution(object):
 
     __repr__ = __str__
 
+    def __eq__(self, other):
+        return float(self) == other
+
+    def __ne__(self, other):
+        return float(self) == other
+
+    def __lt__(self, other):
+        return float(self) == other
+
+    def __le__(self, other):
+        return float(self) == other
+
+    def __gt__(self, other):
+        return float(self) == other
+
+    def __ge__(self, other):
+        return float(self) == other
+
     def __sub__(self, other):
         return float(self) - other
 
@@ -655,7 +655,6 @@ _inflect_engine = inflect.engine()
 
 
 class ConfigReader:
-
     @classmethod
     def _parse_children_config(cls, children, scope):
         """Parses children in the given scope"""
@@ -668,11 +667,15 @@ class ConfigReader:
                 if class_name is None:
                     class_name = stringcase.pascalcase(k)
                 the_class = get_class(class_name=class_name, scope=scope)
-                obj, obj_children = the_class.from_config(config=v, ignore_children=False, try_parsing_children=True
-                                                                        , scope=scope)
+                obj, obj_children = the_class.from_config(
+                    config=v,
+                    ignore_children=False,
+                    try_parsing_children=True,
+                    scope=scope,
+                )
                 if obj_children is not None and len(obj_children) > 0:
                     remaining_children[k] = obj_children
-                setter_name = 'set_' + k
+                setter_name = "set_" + k
                 if hasattr(cls, setter_name):
                     setters.append((setter_name, obj))
                 else:
@@ -683,12 +686,16 @@ class ConfigReader:
                     singular = k
                 if class_name is None:
                     class_name = stringcase.pascalcase(singular)
-                setter_name = 'set_' + k
+                setter_name = "set_" + k
                 objs = []
                 for current in list(v):
                     the_class = get_class(class_name=class_name, scope=scope)
-                    obj = the_class.from_config(config=current, ignore_children=True
-                                                              , try_parsing_children=True, scope=scope)
+                    obj = the_class.from_config(
+                        config=current,
+                        ignore_children=True,
+                        try_parsing_children=True,
+                        scope=scope,
+                    )
                     objs.append(obj)
                 if hasattr(cls, setter_name):
                     setters.append((setter_name, objs))
@@ -710,14 +717,16 @@ class ConfigReader:
               the rest will be ignored. This can be used to add comments
 
         """
-        keys = key.split(':')
+        keys = key.split(":")
         if len(keys) == 1:
             return keys[0], None
         else:
             return keys[0], keys[1]
 
     @classmethod
-    def read_config(cls, config: Union[str, dict], section: str = None) -> Dict[str, Any]:
+    def read_config(
+        cls, config: Union[str, dict], section: str = None
+    ) -> Dict[str, Any]:
         """
         Reads the configuration from a file or a dict and prepares it for parsing
 
@@ -739,41 +748,45 @@ class ConfigReader:
                 return os.path.exists(nm) and not os.path.isdir(nm)
 
             if not exists(config):
-                name = pathlib.Path('./') / pathlib.Path(config)
+                name = pathlib.Path("./") / pathlib.Path(config)
                 if exists(name):
                     config = str(name.absolute())
                 else:
-                    name = (pathlib.Path('./.negmas') / config).absolute()
+                    name = (pathlib.Path("./.negmas") / config).absolute()
                     if exists(name):
                         config = str(name)
                     else:
-                        name = (pathlib.Path(os.path.expanduser('~/.negmas')) / config).absolute()
+                        name = (
+                            pathlib.Path(os.path.expanduser("~/.negmas")) / config
+                        ).absolute()
                         if exists(name):
                             config = str(name)
                         else:
-                            raise ValueError(f'Cannot find config in {config}.')
-            with open(config, 'r') as f:
-                if config.endswith('.json'):
+                            raise ValueError(f"Cannot find config in {config}.")
+            with open(config, "r") as f:
+                if config.endswith(".json"):
                     config = json.load(f)
-                elif config.endswith('.cfg'):
+                elif config.endswith(".cfg"):
                     config = eval(f.read())
-                elif config.endswith('.yaml') or config.endswith('.yml'):
+                elif config.endswith(".yaml") or config.endswith(".yml"):
                     config = yaml.safe_load(f)
                 else:
-                    raise ValueError(f'Cannot parse {config}')
+                    raise ValueError(f"Cannot parse {config}")
 
         if section is not None:
             config = config[section]  # type: ignore
 
-        return config # type: ignore
+        return config  # type: ignore
 
     @classmethod
-    def from_config(cls
-                    , config: Union[str, dict]
-                    , section: str = None
-                    , ignore_children: bool = True
-                    , try_parsing_children: bool = True
-                    , scope=None):
+    def from_config(
+        cls,
+        config: Union[str, dict],
+        section: str = None,
+        ignore_children: bool = True,
+        try_parsing_children: bool = True,
+        scope=None,
+    ):
         """
         Creates an object of this class given the configuration info
 
@@ -813,29 +826,44 @@ class ConfigReader:
 
         # now we have a dict called config which has our configuration
 
-        myconfig = {}   # parts of the config that can directly be parsed
-        children = {}   # parts of the config that need further parsing
-        setters = []  # the setters are those configs that have a set_ function for them.
+        myconfig = {}  # parts of the config that can directly be parsed
+        children = {}  # parts of the config that need further parsing
+        setters = (
+            []
+        )  # the setters are those configs that have a set_ function for them.
 
         def _is_simple(x):
             """Tests whether the input can directly be parsed"""
-            return x is None or isinstance(x, int) or isinstance(x, str) or isinstance(x, float) or \
-                   (isinstance(x, Iterable) and not isinstance(x, dict) and all(_is_simple(_) for _ in list(x)))
+            return (
+                x is None
+                or isinstance(x, int)
+                or isinstance(x, str)
+                or isinstance(x, float)
+                or (
+                    isinstance(x, Iterable)
+                    and not isinstance(x, dict)
+                    and all(_is_simple(_) for _ in list(x))
+                )
+            )
 
         def _set_simple_config(key, v) -> Optional[Dict[str, Any]]:
             """Sets a simple value v for key taken into accout its class and the class we are constructing"""
             key_name, class_name = cls._split_key(key)
-            _setter = 'set_' + key_name
+            _setter = "set_" + key_name
             params = {}
             if hasattr(cls, _setter):
                 setters.append((_setter, v))
                 return None
-            params[key_name] = v if class_name is None else get_class(class_name=class_name, scope=scope)(v)
+            params[key_name] = (
+                v
+                if class_name is None
+                else get_class(class_name=class_name, scope=scope)(v)
+            )
             return params
 
         # read the configs key by key and try to parse anything that is simple enough to parse
 
-        for k, v in config.items(): # type: ignore
+        for k, v in config.items():  # type: ignore
             if isinstance(v, Dict):
                 children[k] = v
             elif isinstance(v, Iterable) and not isinstance(v, str):
@@ -847,7 +875,7 @@ class ConfigReader:
                     if val is not None:
                         myconfig.update(val)
                 else:
-                    children[k] = v # type: ignore
+                    children[k] = v  # type: ignore
             else:
                 # that is a simple value of the form k:class_name = v. We construct class_name (if it exists) with v
                 val = _set_simple_config(k, v)
@@ -858,17 +886,20 @@ class ConfigReader:
 
         if len(children) > 0 and try_parsing_children:
             if scope is None:
-                ValueError(f'scope is None but that is not allowed. You must pass scope=globals() or scope=locals() to '
-                           f'from_config. If your classes are defined in the global scope pass globals() and if they '
-                           f'are defined in local scope then pass locals(). You can only pass scope=None if you are '
-                           f'sure that all of the constructor parameters of the class you are creating are simple '
-                           f'values like ints floats and strings.')
-            parsed_conf, remaining_children, setters = cls._parse_children_config(children=children
-                                                                                  , scope=scope)
+                ValueError(
+                    f"scope is None but that is not allowed. You must pass scope=globals() or scope=locals() to "
+                    f"from_config. If your classes are defined in the global scope pass globals() and if they "
+                    f"are defined in local scope then pass locals(). You can only pass scope=None if you are "
+                    f"sure that all of the constructor parameters of the class you are creating are simple "
+                    f"values like ints floats and strings."
+                )
+            parsed_conf, remaining_children, setters = cls._parse_children_config(
+                children=children, scope=scope
+            )
             myconfig.update(parsed_conf)
             children = remaining_children
 
-        main_object = cls(**myconfig) # type: ignore
+        main_object = cls(**myconfig)  # type: ignore
 
         if try_parsing_children:
             # we will only have setters if we have children
@@ -890,11 +921,15 @@ class Proxy:
         return getattr(self._obj, item)
 
 
-def get_full_type_name(t: Union[Type[Any], Callable]) -> str:
-    """Gets the ful typename of a type. You *should not* pass an instance to this function but it may just work."""
-    if not hasattr(t, '__module__') and not hasattr(t, '__name__'):
+def get_full_type_name(t: Union[Type[Any], Callable, str]) -> str:
+    """Gets the ful typename of a type. You *should not* pass an instance to this function but it may just work.
+
+    An exception is that if the input is of type `str` or if it is None, it will be returned as it is"""
+    if t is None or isinstance(t, str):
+        return t
+    if not hasattr(t, "__module__") and not hasattr(t, "__name__"):
         t = type(t)
-    return t.__module__ + '.' + t.__name__
+    return t.__module__ + "." + t.__name__
 
 
 def import_by_name(full_name: str) -> Any:
@@ -902,35 +937,43 @@ def import_by_name(full_name: str) -> Any:
     if not isinstance(full_name, str):
         return full_name
     modules: List[str] = []
-    parts = full_name.split('.')
+    parts = full_name.split(".")
     modules = parts[:-1]
-    module_name = '.'.join(modules)
-    item_name = parts[:-1]
+    module_name = ".".join(modules)
+    item_name = parts[-1]
     if len(modules) < 1:
-        raise ValueError(f'Cannot get the object {item_name} in module {module_name}  (modules {modules})')
+        raise ValueError(
+            f"Cannot get the object {item_name} in module {module_name}  (modules {modules})"
+        )
     module = importlib.import_module(module_name)
     return getattr(module, item_name)
 
 
-def get_class(class_name: Union[str, Type], module_name: str = None, scope: dict = None) -> Type:
+def get_class(
+    class_name: Union[str, Type], module_name: str = None, scope: dict = None
+) -> Type:
     """Imports and creates a class object for the given class name"""
     if not isinstance(class_name, str):
         return class_name
     modules: List[str] = []
     if module_name is not None:
-        modules = module_name.split('.')
-    modules += class_name.split('.')
+        modules = module_name.split(".")
+    modules += class_name.split(".")
     if len(modules) < 1:
-        raise ValueError(f'Cannot get the class {class_name} in module {module_name}  (modules {modules})')
+        raise ValueError(
+            f"Cannot get the class {class_name} in module {module_name}  (modules {modules})"
+        )
     class_name = stringcase.pascalcase(modules[-1])
     if len(modules) < 2:
         return eval(class_name, scope)
-    module_name = '.'.join(modules[:-1])
+    module_name = ".".join(modules[:-1])
     module = importlib.import_module(module_name)
     return getattr(module, class_name)
 
 
-def instantiate(class_name: Union[str, Type], module_name: str = None, scope: dict = None, **kwargs) -> Any:
+def instantiate(
+    class_name: Union[str, Type], module_name: str = None, scope: dict = None, **kwargs
+) -> Any:
     """Imports and instantiates an object of a class"""
     return get_class(class_name, module_name)(**kwargs)
 
@@ -959,3 +1002,141 @@ def humanize_time(secs, align=False, always_show_all_units=False):
             else:
                 parts.append("%2d%s%s" % (n, unit, ""))
     return ":".join(parts)
+
+
+class NpEncoder(json.JSONEncoder):
+    def default(self, obj):
+        if isinstance(obj, np.integer):
+            return int(obj)
+        elif isinstance(obj, np.floating):
+            return float(obj)
+        elif isinstance(obj, np.ndarray):
+            return obj.tolist()
+        else:
+            return super(NpEncoder, self).default(obj)
+
+
+def dump(d: Any, file_name: Union[str, os.PathLike, pathlib.Path]) -> None:
+    """
+    Saves an object depending on the extension of the file given. If the filename given has no extension,
+    `DEFAULT_DUMP_EXTENSION` will be used
+
+    Args:
+        d: Object to save
+        file_name: file name
+
+    Remarks:
+
+        - Supported formats are json, yaml
+        - If None is given, the file will be created but will be empty
+        - Numpy arrays will be converted to lists before being dumped
+
+    """
+    file_name = pathlib.Path(file_name).expanduser().absolute()
+    if file_name.suffix == "":
+        file_name = pathlib.Path(str(file_name) + "." + DEFAULT_DUMP_EXTENSION)
+
+    if d is None:
+        with open(file_name, "w") as f:
+            pass
+    if file_name.suffix == ".json":
+        with open(file_name, "w") as f:
+            json.dump(d, f, sort_keys=True, indent=2, cls=NpEncoder)
+    elif file_name.suffix == ".yaml":
+        with open(file_name, "w") as f:
+            yaml.safe_dump(d, f)
+    elif file_name.suffix == ".pickle":
+        with open(file_name, "wb") as f:
+            pickle.dump(d, f)
+    elif file_name.suffix == ".csv":
+        if not isinstance(d, pd.DataFrame):
+            try:
+                d = pd.DataFrame(d)
+            except Exception as e:
+                raise ValueError(f"Failed to convert to a dataframe: {str(e)}")
+        d.to_csv(file_name)
+    else:
+        raise ValueError(f"Unknown extension {file_name.suffix} for {file_name}")
+
+
+def load(file_name: Union[str, os.PathLike, pathlib.Path]) -> Any:
+    """
+    Loads an object depending on the extension of the file given. If the filename given has no extension,
+    `DEFAULT_DUMP_EXTENSION` will be used
+
+    Args:
+        file_name: file name
+
+    Remarks:
+
+        - Supported formats are json, yaml
+        - If None is given, the file will be created but will be empty
+
+    """
+    file_name = pathlib.Path(file_name).expanduser().absolute()
+    if file_name.suffix == "":
+        file_name = pathlib.Path(str(file_name) + "." + DEFAULT_DUMP_EXTENSION)
+    d = {}
+    if file_name.suffix == ".json":
+        with open(file_name, "r") as f:
+            d = json.load(f)
+    elif file_name.suffix == ".yaml":
+        with open(file_name, "r") as f:
+            yaml.safe_load(f)
+    elif file_name.suffix == ".pickle":
+        with open(file_name, "rb") as f:
+            d = pickle.load(f)
+    elif file_name.suffix == ".csv":
+        d = pd.read_csv(file_name).to_dict()
+    else:
+        raise ValueError(f"Unknown extension {file_name.suffix} for {file_name}")
+    return d
+
+
+def add_records(
+    file_name: Union[str, os.PathLike], data: Any, col_names: Optional[List[str]] = None
+) -> None:
+    """
+    Adds records to a csv file
+
+    Args:
+
+        file_name: file name
+        data: data to use for creating the record
+        col_names: Names in the data.
+
+    Returns:
+
+        None
+
+    """
+    data = pd.DataFrame(data=data, columns=col_names)
+    if len(data) < 1:
+        return
+    file_name = pathlib.Path(file_name)
+    file_name.parent.mkdir(parents=True, exist_ok=True)
+    new_file = True
+    mode = "a"
+    if file_name.exists():
+        new_file = False
+        with open(file_name, "r") as f:
+            header = f.readline().strip().strip("\n")
+        cols = header.split(",")
+        for col in cols:
+            if len(col) > 0 and col not in data.columns:
+                data[col] = None
+        if set([_ for _ in data.columns]) == set(cols):
+            data = data.loc[:, cols]
+        else:
+            try:
+                old_data = pd.read_csv(file_name, index_col=None)
+                data = pd.concat((old_data, data), axis=0, ignore_index=True)
+            except Exception as e:
+                print(
+                    f"Failed to read data from file {str(file_name)} will override it"
+                )
+
+            mode = "w"
+            new_file = True
+
+    data.to_csv(str(file_name), index=False, index_label="", mode=mode, header=new_file)
