@@ -484,6 +484,7 @@ class AspirationNegotiator(SAONegotiator, AspirationMixin):
         )
         self.__last_offer_util, self.__last_offer = float("inf"), None
         self.n_outcomes_to_force_presort = 10000
+        self.n_trials = 1
 
     def on_ufun_changed(self):
         super().on_ufun_changed()
@@ -865,7 +866,7 @@ class NaiveTitForTatNegotiator(SAONegotiator):
         )
         return indx
 
-    def propose(self, state: MechanismState) -> Optional[Outcome]:
+    def propose(self, state: MechanismState) -> Optional["Outcome"]:
         indx = self._propose(state)
         self.proposed_utility = self.ordered_outcomes[indx][0]
         return self.ordered_outcomes[indx][1]
